/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *	Side Effects Software Inc
 *	477 Richmond Street West
 *	Toronto, Ontario
 *	Canada   M5V 3E7
 *	416-504-9876
 *
 * NAME:	IMG_Raw.C ( IMG Library, C++)
 *
 * COMMENTS:	Read/Write raw files
 */

#include <sys/types.h>
#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <limits.h>
#include <UT/UT_Endian.h>		// For byte swapping
#include <UT/UT_DSOVersion.h>
#include <UT/UT_SysClone.h>
#include <IMG/IMG_Format.h>
#include "IMG_Raw.h"

#define MAGIC		0x1234567a
#define MAGIC_SWAP	0xa7654321		// Swapped magic number

typedef struct {
    unsigned int	magic;		// Magic number
    unsigned int	xres;		// Width of image
    unsigned int	yres;		// Height of image
    unsigned int	model;		// My color model
    unsigned int	data;		// Data type
} IMG_RawHeader;

class IMG_RawFormat : public IMG_Format {
public:
	     IMG_RawFormat() {}
    virtual ~IMG_RawFormat() {}

    virtual const char	*getFormatName() const;
    virtual const char	*getFormatLabel() const;
    virtual const char	*getFormatDescription() const;
    virtual const char	*getDefaultExtension() const;
    virtual IMG_File	*createFile()	 const;

    // Methods to determine if this is one of our recognized files.
    //	The extension is the first try.  If there are multiple matches,
    //	then we resort to the magic number (when reading)
    virtual int		 checkExtension(const char *filename) const;
    virtual int		 checkMagic(unsigned int) const;

    // Configuration information for the format
    virtual unsigned		getMaxImages() const;
    virtual void		getMaxResolution(unsigned &x,
					         unsigned &y) const;

    virtual int			isReadRandomAccess() const  { return 0; }
    virtual int			isWriteRandomAccess() const { return 0; }
};

const char *
IMG_RawFormat::getFormatName() const
{
    // Very brief label (no spaces)
    return "Raw";
}

const char *
IMG_RawFormat::getFormatLabel() const
{
    // A simple description of the format
    return "Simple Raw Data";
}

const char *
IMG_RawFormat::getFormatDescription() const
{
    // A more verbose description of the image format.  Things you might put in
    // here are the version of the format, etc.
    return "Sample image format.  Not very useful";
}

const char *
IMG_RawFormat::getDefaultExtension() const
{
    // The default extension for the format files.  If there is no default
    // extension, the format won't appear in the menus to choose image format
    // types.
    return "raw";
}

IMG_File *
IMG_RawFormat::createFile() const
{
    return new IMG_Raw;
}

int
IMG_RawFormat::checkExtension(const char *filename) const
{
    static const char	*extensions[] = { ".raw", ".RAW", 0 };
    return matchExtensions(filename, extensions);
}

int
IMG_RawFormat::checkMagic(unsigned int magic) const
{
    // Check if we hit our magic number
    return (magic == MAGIC || magic == MAGIC_SWAP);
}

unsigned
IMG_RawFormat::getMaxImages() const
{
    return 1;
}

void
IMG_RawFormat::getMaxResolution(unsigned &x, unsigned &y) const
{
    x = UINT_MAX;		// Stored as shorts
    y = UINT_MAX;
}


//////////////////////////////////////////////////////////////////
//
//  Raw file loader/saver
//
//////////////////////////////////////////////////////////////////

IMG_Raw::IMG_Raw()
{
    myByteSwap = 0;
}

IMG_Raw::~IMG_Raw()
{
    close();
}

int
IMG_Raw::open()
{
    return readHeader();
}

int
IMG_Raw::create(const IMG_Stat &stat)
{
    myStat = stat;

#if 0
    // This format can handle everything the IMG library can throw at it.
    // However, you may need to uncomment the following code for a different
    // format
    if (myStat.getDataType() != IMG_UCHAR && myStat.getDataType() != IMG_USHORT)
    {
	if (!setTranslator(IMG_UCHAR, myStat.getColorModel()))
	    return 0;
    }
#endif

    return writeHeader();
}

int
IMG_Raw::closeFile()
{
    // If we're writing data, flush out the stream
    if (myOS) myOS->flush();	// Flush out the data

    return 1;	// return success
}

static inline void
swapHeader(IMG_RawHeader &header)
{
    UTswapBytes((int *)&header, sizeof(header));
}

int
IMG_Raw::readHeader()
{
    IMG_RawHeader	header;

    if (!myIS->read((char *)&header, sizeof(IMG_RawHeader)))
	return 0;

    if (header.magic == MAGIC_SWAP)
    {
	myByteSwap = 1;
	swapHeader(header);
    }
    else if (header.magic != MAGIC)
	return 0;			// Magic number failed.

    myStat.setResolution(header.xres, header.yres);
    myStat.setColorModel((IMG_ColorModel)header.model);
    myStat.setDataType((IMG_DataType)header.data);
    myStat.setImageCount(1);

    // Now, we're ready to read the data.
    return 1;
}

int
IMG_Raw::writeHeader()
{
    IMG_RawHeader	header;

    header.magic = MAGIC;		// Always create native byte order
    header.xres = myStat.getXres();
    header.yres = myStat.getYres();
    header.model = myStat.getColorModel();
    header.data  = myStat.getDataType();

    if (!myOS->write((char *)&header, sizeof(IMG_RawHeader)))
	return 0;

    // Now, we're ready to write the scanlines...
    return 1;
}

int
IMG_Raw::readScanline(int y, void *buf)
{
    int		nbytes;

    if (y >= myStat.getYres()) return 0;

    nbytes = myStat.bytesPerScanline();
    if (!myIS->read((char *)buf, nbytes))
	return 0;

    // If the file was written on a different architecture, we might need to
    // swap the data.
    if (myByteSwap)
    {
	switch (myStat.getDataType())
	{
	    case IMG_UCHAR:	break;		// Nope
	    case IMG_USHORT:	UTswapBytes((short *)buf, nbytes/sizeof(short));
				break;
	    case IMG_UINT:	UTswapBytes((int *)buf, nbytes/sizeof(int));
				break;
	    case IMG_FLOAT:	UTswapBytes((float *)buf, nbytes/sizeof(float));
				break;
	}
    }

    return 1;
}

int
IMG_Raw::writeScanline(int /*y*/, const void *buf)
{
    // If we specified a translator in creation, the buf passed in will be in
    // the format we want, that is, the translator will make sure the data is
    // in the correct format.

    // Since we always write in native format, we don't have to swap
    return (!myOS->write((char *)buf, myStat.bytesPerScanline())) ? 0 : 1;
}

////////////////////////////////////////////////////////////////////
//
//  Now, we load the format
//
////////////////////////////////////////////////////////////////////
void
newIMGFormat(void *)
{
    new IMG_RawFormat();
}
