/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *      Side Effects Software Inc
 *      477 Richmond Street West
 *      Toronto, Ontario
 *      Canada   M5V 3E7
 *      416-504-9876
 *
 * NAME:        OBJ_Shake.C (Custom Library, C++)
 *
 * COMMENTS:    This custom object defines a geometry object that
 *              has parameters which define a frame-dependent jitter
 *              in its translation. This can be used to make earth-quake
 *              type effects. No doubt next year will be the year of the
 *              Earthquake movie.
 */

#include <UT/UT_DSOVersion.h>
#include <UT/UT_Math.h>
#include <CH/CH_LocalVariable.h>
#include <PRM/PRM_Include.h>
#include <OP/OP_Director.h>
#include <OP/OP_Operator.h>
#include <OP/OP_OperatorTable.h>
#include <OBJ/OBJ_Node.h>
#include "OBJ_Shake.h"



// shakeIndirect is an array of indices that are used to access parameters
// quickly. The indices are initialized by the EVAL_FLOAT methods in the
// header file.
int *OBJ_Shake::shakeIndirect = 0;



// Constructor for new object class
OBJ_Shake::OBJ_Shake(OP_Network *net, const char *name, OP_Operator *op)
	  :OBJ_Geometry(net, name, op)
{
    // initialize local member data here.

    // shakeIndirect is initialized to an array of 3 indices (each initialized
    // to -1 -- ie not accessed yet)
    if (!shakeIndirect) shakeIndirect = allocIndirect(3);
}



// virtual destructor for new object class
OBJ_Shake::~OBJ_Shake()
{
}



// here is the name of the parameter that is used by the shake object
static PRM_Name OBJjitter("jitter",           "Jitter scale");



// this is the template list that defines the new parameters that are
// used by the shake object.
static PRM_Template	templatelist[] =
{
    // Here we define the new parameter
    PRM_Template(PRM_XYZ_J, 3, &OBJjitter),

    // followed by this blank terminating Template.
    PRM_Template()
};



// this function returns the OP_TemplatePair that combines the parameters
// of this object with those of its ancestors in the (object type hierarchy)

OP_TemplatePair	*
OBJ_Shake::buildTemplatePair(OP_TemplatePair *prevstuff)
{
    OP_TemplatePair	*shake, *geo;

    // Here, we have to "inherit" template pairs from geometry and beyond.  To
    // do this, we first need to instantiate our template list, then add the
    // base class templates.
    shake = new OP_TemplatePair(templatelist, prevstuff);
    geo = new OP_TemplatePair(OBJ_Geometry::getTemplateList(), shake);
    return geo;
}



// the myConstructor method is used to create new objects of the correct
// type from the OperatorTable.
OP_Node *
OBJ_Shake::myConstructor(OP_Network *net, const char *name, OP_Operator *op)
{
    return new OBJ_Shake(net, name, op);
}



// this method pre-multiplies the given matrix with an appropriate transform
// to shake the object. our shake transform gets inserted into the object
// transform pipeline essentially before the transform parameters.
int
OBJ_Shake::applyInputIndependentTransform(OP_Context &context, UT_DMatrix4 &mat)
{
    const float	 t = context.myTime;	// extract time from OP_Context
    float 	 jx, jy, jz;
    unsigned	 seed;
    int		 modified;

    // call OBJ_Geometry::applyInputIndependentTransform() so that we don't
    // lose any information
    modified = OBJ_Geometry::applyInputIndependentTransform(context, mat);
    if (error() >= UT_ERROR_ABORT)
    {
	// don't do anything since an error has occurred.
	return modified;
    }

    // first we compute our jitter values as a random value within
    // the given jitter scale using a frame dependent seed value
    // (different for x, y, and z)
    jx = JX(t); jy = JY(t); jz = JZ(t);

    seed = (int)OPgetDirector()->getChannelManager()->getSample(t);
    jx *= 2*UTfastRandom(seed) - 1.0;

    seed ^= 0xdeadbeef;
    jy *= 2*UTfastRandom(seed) - 1.0;

    seed ^= 0xfadedcab;
    jz *= 2*UTfastRandom(seed) - 1.0;

    // we add our jitter to the object transform
    mat.pretranslate(jx, jy, jz);

    // return 1 to indicate that we have modified the input matrix.
    // if we didn't modify mat, then we should return 0 instead.
    return 1;
}



OP_ERROR
OBJ_Shake::cookMe(OP_Context &context)
{
    OP_ERROR	errorstatus;

    // OBJ_Geometry::cookMe computes the local and global transform, and
    // also the inverse of the global transform matrix. These are stored
    // in myXform, myWorldXform, and myIWorldXform, respectively.
    errorstatus = OBJ_Geometry::cookMe(context);

    // since the jitter value is frame dependent and this object won't
    // be flagged as time dependent through any other method, we explicitly
    // flag it here. 
    // NB: this flag gets reset at the beginning of every cook!
    ourFlags.timeDep = 1;

    return errorstatus;
}



// this function installs the new object in houdini's object table.
void
newObjectOperator(OP_OperatorTable *table)
{
    table->addOperator(new OP_Operator("shake", "Shake",
		       OBJ_Shake::myConstructor,
		       OBJ_Shake::buildTemplatePair(0), 0, 1,
		       0));
}
