/*
 * PROPRIETARY INFORMATION.  This software is proprietary to
 * Side Effects Software Inc., and is not to be reproduced,
 * transmitted, or disclosed in any way without written permission.
 *
 * Produced by:
 *      Side Effects
 *      477 Richmond Street West
 *      Toronto, Ontario
 *      Canada   M5V 3E7
 *      416-504-9876
 */

#include <string.h>
#include <malloc.h>
#include <math.h>
#include <UT/UT_DSOVersion.h>
#include <EXPR/EXPR.h>
#include <EXPR/EX_Vector.h>
#include <OP/OP_Director.h>
#include <OP/OP_Operator.h>
#include <CH/CH_Support.h>
#include <CMD/CMD_Manager.h>


//
// This is a utility function to get the current operator which
//	is being evaluated.
//
static OP_Node *
getCwd(float &time)
{
    OP_Channels	*chp;
    CH_Manager	*chman;

    chman = OPgetDirector()->getChannelManager();
    time = chman->getEvaluateTime();
    chp = (OP_Channels *)chman->getEvalCollection();
    if (!chp) chp = OPgetDirector()->getCwd()->getChannels();

    return (chp) ? chp->getNode() : 0;
}

//
// This is a function which will find a node from our current node's
//	location.
//
static OP_Node *
findOp(const char *object, OP_InterestType interest_type = OP_INTEREST_NAME)
{
    OP_Node	*cwd;		// Where to search from
    OP_Node	*here;		// Where I currently am
    OP_Node	*src;
    OP_Context	 context;

    // Find our current evaluation node
    here = getCwd(context.myTime);
    if (!here) return 0;

    // If there's a full path specification, search from the top
    cwd = (*object == '/') ? OPgetDirector() : here;

    if (*object == '\0' || !strcmp(object, "."))
	 src = (OP_Node *)cwd;
    else src = (OP_Node *)cwd->findNode(object);

    // Check to make sure we found the node.
    if (!src) return 0;

    // We need to be told when certain things happen to the node we just found.
    // For example, if the node gets re-named, we need to be told so that the
    // expression can fix itself.  As well, in many cases, we'll also want to
    // be notified when the data changes.  These types of interests can be
    // found in OP/OP_DataTypes.h
    here->addExtraInput(src, interest_type);

    return src;
}

// Callback function to evaluate the max of two numbers
static void
fn_max(EV_FUNCTION *, EV_SYMBOL *result, EV_SYMBOL **argv, int)
{
    if (argv[0]->value.fval > argv[1]->value.fval)
	 result->value.fval = argv[0]->value.fval;
    else result->value.fval = argv[1]->value.fval;
}

// Callback function to evaluate the minimum of two numbers
static void
fn_min(EV_FUNCTION *, EV_SYMBOL *result, EV_SYMBOL **argv, int)
{
    if (argv[0]->value.fval > argv[1]->value.fval)
	 result->value.fval = argv[1]->value.fval;
    else result->value.fval = argv[0]->value.fval;
}

// This callback is a little more tricky, it evaluates the type of the
//	operator specified.
static void
fn_optype(EV_FUNCTION *, EV_SYMBOL *result, EV_SYMBOL **argv, int)
{
    OP_Node		*node;		// Node specified
    const OP_Operator	*opdef;		// The operator definition

    if (node = findOp(argv[0]->value.sval))
    {
	// Get the operator table definition which stores the type of operator
	opdef = node->getOperator();

	// Set the result to be a duplication of the operator table.  Make
	//	to allocate memory for a string result.
	result->value.sval = strdup(opdef->getName());
    }
    else result->value.sval = 0;	// An empty string
}

static void
fn_vectorsum(EV_FUNCTION *, EV_SYMBOL *result, EV_SYMBOL **argv, int)
{
    ev_Vector	*v0 = (ev_Vector *)argv[0]->value.data;
    ev_Vector	*v1 = (ev_Vector *)argv[1]->value.data;
    ev_Vector	*sum  = (ev_Vector *)result->value.data;

    sum->copy(*v0);
    sum->add(*v1);
}

// A couple of defines to make life a lot easier for us
#define EVF	EV_TYPEFLOAT
#define EVS	EV_TYPESTRING
#define EVV	EV_TYPEVECTOR

static int	floatArgs[] = { EVF, EVF };
static int	stringArgs[] = { EVS };
static int	vectorArgs[] = { EVV, EVV };

static EV_FUNCTION funcTable[] = {
    EV_FUNCTION(0, "max",	2, EVF,	floatArgs,	fn_max),
    EV_FUNCTION(0, "min",	2, EVF,	floatArgs,	fn_min),
    EV_FUNCTION(0, "optype",	1, EVS,	stringArgs,	fn_optype),
    EV_FUNCTION(0, "vectorsum",	2, EVV,	vectorArgs,	fn_vectorsum),
    EV_FUNCTION(),
};

void
CMDextendLibrary(CMD_Manager *)
{
    int		i;

    for (i = 0; funcTable[i].getName(); i++)
	ev_AddFunction(&funcTable[i]);
}
